==> lneural-0.1.0.tar.bz2: lneural-0.1.0/network.lua <==
local Net   = {}
local NetMT = {__index = Net}

local Act = require('lneural.activation')
local S   = require('lneural.serializer')

math.randomseed( os.time() )

function Net.new(...)
    local self = {}
    setmetatable(self, NetMT)
    
    self.n_neurons  = {...}
    self.n_layers   = #self.n_neurons
    self.acc        = {}
    
    -- add a bias neuron to the input and hidden layers 
    local nn        = self.n_neurons
    for l = 1, self.n_layers - 1 do nn[l] = nn[l] + 1 end
    
    -- calculate the total number of neurons and connections
    local tneurons, tconn   = 0, 0
    
    for l = 1, self.n_layers do
        tneurons = tneurons + nn[l]
        if(l > 1) then tconn = tconn + nn[ l - 1 ] * nn[ l ] end
    end
    
    self.total_neurons, self.total_conn = tneurons, tconn
    
    -- calculate the initial weights and the indexes
    self:randomize_weights()
    self.conn_idx, self.neuron_idx = self:calc_idx()
    
    -- parameters
    self.eta                = 0.7
    self.eta_min            = 0.002
    self.increase_factor    = 1.03
    self.decrease_factor    = 0.95
    self:set_activation_function('sigmoid')
    
    return self
end

function Net:randomize_weights()
    local w = {}
    for n = 1, self.total_conn do w[n] = math.random() - 0.5 end
    self.weights = w
end

-- The connections are stored in single-dimension array; this function 
-- calculates the index of the first connection of a neuron and the index of
-- the first neuron of a layer.
function Net:calc_idx()
    local conn_idx, neuron_idx  = {}, {1}
    local neuron, idx           = 1 , 1

    for l = 2, self.n_layers do
        local n_this, n_next    = self.n_neurons[l - 1], self.n_neurons[ l ]
    
        for n = 1, n_this do
            conn_idx[ neuron ] = idx
            neuron = neuron + 1
            idx = idx + n_next
        end
    
        neuron_idx[ l ] = neuron
    end
    
    return conn_idx, neuron_idx
end

-- returns the neuron id of a neuron identified by it's layer and it's position
-- in that layer.
function Net:get_neuron(layer, neuron)
    return self.neuron_idx[ layer ] + neuron - 1
end

-- returns the index of the connection, the number of conenctions and the neuron
-- index, of a neuron identified by it's layer and it's position in the layer
function Net:get_weights(layer, neuron)
    local neuron = self:get_neuron(layer, neuron)
    return self.conn_idx[ neuron ], self.n_neurons[ layer + 1 ] - 1, neuron
end

function Net:run(input, offset, out)
    local n_neurons, weights, acc = self.n_neurons, self.weights, self.acc
    local to, offset = 1, offset or 0
    
    for i = 1, n_neurons[1] do acc[i] = input[i + offset] end       -- input
    for i = n_neurons[1] + 1, self.total_neurons do acc[i] = 0 end  -- others
    
    for l = 1, self.n_layers - 1 do
        to = to + n_neurons[l]
        acc[ to - 1 ] = 1       -- the bias always gives 1
        
        for n = 1, n_neurons[l] do
            local conn, offset, from = self:get_weights(l, n)
            
            for o = 0, offset do
                local lto = to + o
                acc[lto] = acc[lto] + weights[conn + o] * acc[from]
            end
        end
        
        local l = to + n_neurons[l + 1] - 1
        for i = to, l do acc[i] = self:activation( acc[i] ) end
    end
    
    out = out or {}
    local n = self.neuron_idx[ self.n_layers ] - 1
    for i = 1, n_neurons[ self.n_layers ] do out[i] = acc[n + i] end
    
    return out
end

function Net:delta_output(delta, data, out, out_idx)
    local nout_idx  = self.neuron_idx[ self.n_layers ]
    local mse, a    = 0, self.momentum
            
    for i, val in ipairs(out) do
        local err = data[out_idx + i] - val
        delta[nout_idx + i - 1] = err * self:derivative(val)
                
        err = err * self.msef
        mse = mse + err * err
    end
    
    return mse
end

function Net:delta_hidden(delta)
    local weights, acc, eta  = self.weights, self.acc, self.eta

    for l = self.n_layers - 1, 2, -1 do
        for j = 1, self.n_neurons[l] do
            local sum       = 0
            local conn, offset, from = self:get_weights(l, j)
            local kf        = self.neuron_idx[l + 1]
                    
            for o = 0, offset do
                sum = sum + delta[kf + o] * weights[conn + o]
            end
            
            delta[from] = eta * self:derivative( acc[from] ) * sum
        end
    end
end

function Net:update_weights(delta)
    local weights, acc = self.weights, self.acc

    for l = 1, self.n_layers - 1 do
        for n = 1, self.n_neurons[l] do
            local conn, offset, j = self:get_weights(l, n)
            local kf = self.neuron_idx[l + 1]
                    
            for o = 0, offset do
                local id, to = conn + o, kf + o
                local old = weights[id]
                weights[id] = weights[id] + delta[to] * acc[j]
            end 
        end
    end
end

function Net:train(data, desired_mse, max_epochs, cb)
	local epoch, mse, last_mse  = 0, -1, -1
	local desired_mse           = desired_mse or 0
	local max_epochs            = max_epochs or math.huge
    local data_len              = data:length()
    local delta, out            = {}, {}
    
	while true do
        if cb then
            local cont = cb(epoch, mse)
            if not cont then break end
        end
        
        if mse > last_mse then
            self.eta = self.eta * self.decrease_factor
        else
            self.eta = self.eta * self.increase_factor
        end
        
        if self.eta < self.eta_min then self.eta = self.eta_min end
        last_mse, mse = mse, 0
        
		for set = 1, data_len do
            local offset    = data:get_index(set) - 1
            self:run            ( data, offset, out )
            
            local out_idx = offset + data.n_inp
            mse = mse + 
            self:delta_output   (delta, data, out, out_idx)
            
            self:delta_hidden   ( delta )
            self:update_weights ( delta )
        end
        
        epoch   = epoch + 1
        mse     = mse / data_len
        
        if mse      <  desired_mse  then break end
        if epoch    >= max_epochs   then break end
	end
	
	return mse, epoch
end

function Net:validate(data, verbose, t)
    local out           = {}
    local mse, suc, t   = 0, 0, t or 0.5
    local noutput, len  = self.n_neurons[ self.n_layers ], data:length()
    
    for set = 1, len do
        local offset = data:get_index(set) - 1
        local outpos = offset + data.n_inp
        self:run(data, offset, out)
        
        local tsuc = 0
        
        for n = 1, noutput do
            local d, y = data[ outpos + n ], out[n]
            
            if verbose then 
                print( string.format('S: %d-%d | D: % .6f | Y: % .6f', set, n, d, y) )
            end
            
            local err = d - y
            if math.abs(err) < t then tsuc = tsuc + 1 end
            
            err = err * self.msef
            mse = mse + err * err
        end
        
        if tsuc == noutput then suc = suc + 1 end
    end
    
    mse         = mse / len
    local rate  = suc / len
    
    if verbose then
        print( string.format('MSE: %.6f | Rate: %0.2f%%', mse, rate * 100) )
    end
    
    return mse, rate
end

function Net:set_activation_function(name)
    local act = Act[name]
    
    self.activation = act.activation
    self.derivative = act.derivative
    self.msef       = act.msef or 1
    self.actname    = name
end

function Net:progress_callback(period)
    return function(epoch, mse)
        if epoch % period == 0 then
            local msg = string.format('epoch: %d | eta: %.6f | mse: %.6f', 
                epoch, self.eta, mse)
            print(msg)
        end
        
        return true
    end
end

function Net:save_to_file(path)
    self.acc = {}

    local str   = S(self)
    local f     = io.open(path, 'w')
    f:write('return ', str)
    f:close()
end

function Net.new_from_file(path)
    local self = dofile(path)
    setmetatable(self, NetMT)
    
    self:set_activation_function(self.actname)
    
    return self
end

return Net

==> lneural-0.2.0.tar.bz2: lneural-0.2.0/network.lua <==
--[[
	This file is part of lneural.

	lneural is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	lneural is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with lneural.  If not, see <http://www.gnu.org/licenses/>.
	
	Copyright (C) 2010 Lucas Hermann Negri
--]]

local Net   = {}
local NetMT = {__index = Net}

local Act = require('lneural.activation')
local S   = require('lneural.serializer')

math.randomseed( os.time() )

function Net.new(...)
    local self = {}
    setmetatable(self, NetMT)
    
    self.n_neurons  = {...}
    self.n_layers   = #self.n_neurons
    self.acc        = {}
    
    -- add a bias neuron to the input and hidden layers 
    local nn        = self.n_neurons
    for l = 1, self.n_layers - 1 do nn[l] = nn[l] + 1 end
    
    -- calculate the total number of neurons and connections
    local tneurons, tconn   = 0, 0
    
    for l = 1, self.n_layers do
        tneurons = tneurons + nn[l]
        if(l > 1) then tconn = tconn + nn[ l - 1 ] * nn[ l ] end
    end
    
    self.total_neurons, self.total_conn = tneurons, tconn
    
    -- calculate the initial weights and the indexes
    self:randomize_weights()
    self.conn_idx, self.neuron_idx = self:calc_idx()
    
    -- parameters
    self.eta                = 0.7
    self.eta_min            = 0.002
    self.increase_factor    = 1.03
    self.decrease_factor    = 0.95
    self:set_activation_function('sigmoid')
    
    return self
end

function Net:randomize_weights()
    local w = {}
    for n = 1, self.total_conn do w[n] = math.random() - 0.5 end
    self.weights = w
end

-- The connections are stored in single-dimension array; this function 
-- calculates the index of the first connection of a neuron and the index of
-- the first neuron of a layer.
function Net:calc_idx()
    local conn_idx, neuron_idx  = {}, {1}
    local neuron, idx           = 1 , 1

    for l = 2, self.n_layers do
        local n_this, n_next    = self.n_neurons[l - 1], self.n_neurons[ l ]
    
        for n = 1, n_this do
            conn_idx[ neuron ] = idx
            neuron = neuron + 1
            idx = idx + n_next
        end
    
        neuron_idx[ l ] = neuron
    end
    
    return conn_idx, neuron_idx
end

-- returns the neuron id of a neuron identified by it's layer and it's position
-- in that layer.
function Net:get_neuron(layer, neuron)
    return self.neuron_idx[ layer ] + neuron - 1
end

-- returns the index of the connection, the number of conenctions and the neuron
-- index, of a neuron identified by it's layer and it's position in the layer
function Net:get_weights(layer, neuron)
    local neuron = self:get_neuron(layer, neuron)
    return self.conn_idx[ neuron ], self.n_neurons[ layer + 1 ] - 1, neuron
end

function Net:run(input, offset, out)
    local n_neurons, weights, acc = self.n_neurons, self.weights, self.acc
    local to, offset = 1, offset or 0
    
    for i = 1, n_neurons[1] do acc[i] = input[i + offset] end       -- input
    for i = n_neurons[1] + 1, self.total_neurons do acc[i] = 0 end  -- others
    
    for l = 1, self.n_layers - 1 do
        to = to + n_neurons[l]
        acc[ to - 1 ] = 1       -- the bias always gives 1
        
        for n = 1, n_neurons[l] do
            local conn, offset, from = self:get_weights(l, n)
            
            for o = 0, offset do
                local lto = to + o
                acc[lto] = acc[lto] + weights[conn + o] * acc[from]
            end
        end
        
        local l = to + n_neurons[l + 1] - 1
        for i = to, l do acc[i] = self:activation( acc[i] ) end
    end
    
    out = out or {}
    local n = self.neuron_idx[ self.n_layers ] - 1
    for i = 1, n_neurons[ self.n_layers ] do out[i] = acc[n + i] end
    
    return out
end

function Net:delta_output(delta, data, out, out_idx)
    local nout_idx  = self.neuron_idx[ self.n_layers ]
    local mse       = 0
            
    for i, val in ipairs(out) do
        local err = data[out_idx + i] - val
        delta[nout_idx + i - 1] = err * self:derivative(val)
        mse = mse + err * err
    end
    
    return mse
end

function Net:delta_hidden(delta)
    local weights, acc, eta  = self.weights, self.acc

    for l = self.n_layers - 1, 2, -1 do
        for j = 1, self.n_neurons[l] do
            local sum       = 0
            local conn, offset, from = self:get_weights(l, j)
            local kf        = self.neuron_idx[l + 1]
                    
            for o = 0, offset do
                sum = sum + delta[kf + o] * weights[conn + o]
            end
            
            delta[from] = self:derivative( acc[from] ) * sum
        end
    end
end

function Net:update_weights(delta)
    local weights, acc, eta = self.weights, self.acc, self.eta

    for l = 1, self.n_layers - 1 do
        for n = 1, self.n_neurons[l] do
            local conn, offset, j = self:get_weights(l, n)
            local kf = self.neuron_idx[l + 1]
                    
            for o = 0, offset do
                local id, to = conn + o, kf + o
                weights[id] = weights[id] + eta * delta[to] * acc[j]
            end 
        end
    end
end

function Net:train(data, desired_mse, max_epochs, cb)
	local epoch, mse, last_mse  = 0, -1, -1
	local desired_mse           = desired_mse or 0
	local max_epochs            = max_epochs or math.huge
    local data_len              = data:length()
    local delta, out            = {}, {}
    
	while true do
        if cb then
            local cont = cb(epoch, mse)
            if not cont then break end
        end
        
        if mse > last_mse then
            self.eta = self.eta * self.decrease_factor
        else
            self.eta = self.eta * self.increase_factor
        end
        
        if self.eta < self.eta_min then self.eta = self.eta_min end
        last_mse, mse = mse, 0
        
		for set = 1, data_len do
            local offset    = data:get_index(set) - 1
            self:run            ( data, offset, out )
            
            local out_idx = offset + data.n_inp
            mse = mse + 
            self:delta_output   (delta, data, out, out_idx)
            
            self:delta_hidden   ( delta )
            self:update_weights ( delta )
        end
        
        epoch   = epoch + 1
        mse     = mse / (data_len * 2);
        
        if mse      <  desired_mse  then break end
        if epoch    >= max_epochs   then break end
	end
	
	return mse, epoch
end

function Net:validate(data, verbose, t)
    local out           = {}
    local mse, suc, t   = 0, 0, t or 0.5
    local noutput, len  = self.n_neurons[ self.n_layers ], data:length()
    
    for set = 1, len do
        local offset = data:get_index(set) - 1
        local outpos = offset + data.n_inp
        self:run(data, offset, out)
        
        local tsuc = 0
        
        for n = 1, noutput do
            local d, y = data[ outpos + n ], out[n]
            
            if verbose then 
                print( string.format('S: %d-%d | D: % .6f | Y: % .6f', set, n, d, y) )
            end
            
            local err = d - y
            if math.abs(err) < t then tsuc = tsuc + 1 end
            mse = mse + err * err
        end
        
        if tsuc == noutput then suc = suc + 1 end
    end
    
    mse         = mse / (len*2)
    local rate  = suc / len
    
    if verbose then
        print( string.format('MSE: %.6f | Rate: %0.2f%%', mse, rate * 100) )
    end
    
    return mse, rate
end

function Net:set_activation_function(name)
    local act = Act[name]
    
    self.activation = act.activation
    self.derivative = act.derivative
    self.actname    = name
end

function Net:progress_callback(period)
    return function(epoch, mse)
        if epoch % period == 0 then
            local msg = string.format('epoch: %d | eta: %.6f | mse: %.6f', 
                epoch, self.eta, mse)
            print(msg)
        end
        
        return true
    end
end

function Net:save_to_file(path)
    self.acc = {}

    local str   = S(self)
    local f     = io.open(path, 'w')
    f:write('return ', str)
    f:close()
end

function Net.new_from_file(path)
    local self = dofile(path)
    setmetatable(self, NetMT)
    
    self:set_activation_function(self.actname)
    
    return self
end

return Net
